# Copyright (C) 2016 Midokura SARL.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg


lbaas_group = cfg.OptGroup(name='lbaas',
                           title='Neutron LBaaS Options')


lbaas_opts = [
    cfg.StrOpt('default_listener_protocol', default='HTTP',
               help='Default protocol for Neutron LBaaS Listeners'),
    cfg.StrOpt('default_pool_protocol', default='HTTP',
               help='Default protocol for Neutron LBaaS Pools'),
    cfg.StrOpt('default_health_monitor_protocol', default='HTTP',
               help='Default protocol for Neutron LBaaS Health Monitors'),
    cfg.ListOpt('session_persistence_types',
                default=['SOURCE_IP', 'HTTP_COOKIE', 'APP_COOKIE'],
                help='Supported session persistence types for Neutron LBaaS'),
    cfg.BoolOpt('test_with_noop', default=False,
                help='Configures tempest to run with no-op drivers. Ignores '
                     'the operating_status during tests.'),
]



# Copyright 2016-2017 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import sys

import mock

from neutron_lbaas.tests.unit.db.loadbalancer import test_db_loadbalancerv2
from neutron_lib import context

with mock.patch.dict(
        sys.modules, {'f5lbaasdriver': mock.Mock(__version__="1.0.0"),
                      'f5lbaasdriver.v2': mock.Mock(),
                      'f5lbaasdriver.v2.bigip': mock.Mock(),
                      'f5lbaasdriver.v2.bigip.driver_v2': mock.Mock()}):
    from neutron_lbaas.drivers.f5 import driver_v2


class FakeModel(object):
    def __init__(self, id):
        self.id = id


class DriverTest(object):
    def __init__(self, parent, manager, model, mocked_root):
        self.parent = parent
        self.context = parent.context
        self.driver = parent.driver
        self.manager = manager
        self.model = model
        self.mocked_root = mocked_root

        self.create(model)
        self.update(model, model)
        self.delete(model)

    def create(self, model):
        self.manager.create(self.context, model)
        self.mocked_root.create.assert_called_with(self.context, model)

    def update(self, old_model, model):
        self.manager.update(self.context, old_model, model)
        self.mocked_root.update.assert_called_with(self.context,
                                                   old_model, model)

    def delete(self, model):
        self.manager.delete(self.context, model)
        self.mocked_root.delete.assert_called_with(self.context, model)

    def refresh(self):
        self.manager.refresh(self.context, self.model)
        self.mocked_root.refresh.assert_called_with(self.context, self.model)

    def stats(self):
        self.manager.stats(self.context, self.model)
        self.mocked_root.stats.assert_called_with(self.context, self.model)


class TestF5DriverV2(test_db_loadbalancerv2.LbaasPluginDbTestCase):
    """Unit tests for F5 LBaaSv2 service provider.

    All tests follow the same pattern: call driver method and test
    that driver delegated request to its driver implementation object.

    To test independent of full neutron-lbaas test suite:
        tox -e py27 -- neutron_lbaas.tests.unit.drivers.f5
    """

    def setUp(self):
        super(TestF5DriverV2, self).setUp()
        plugin = mock.Mock()

        self.driver = driver_v2.F5LBaaSV2Driver(plugin)
        self.driver.f5 = mock.Mock()
        self.context = context.get_admin_context()

    def test_load_balancer(self):

        tester = DriverTest(self,
                            self.driver.load_balancer,
                            FakeModel("loadbalancer-01"),
                            self.driver.f5.loadbalancer)

        # additional loadbalancer-only tests
        tester.refresh()
        tester.stats()

    def test_listener(self):
        DriverTest(self,
                   self.driver.listener,
                   FakeModel("listener-01"),
                   self.driver.f5.listener)

    def test_pool(self):
        DriverTest(self,
                   self.driver.pool,
                   FakeModel("pool-01"),
                   self.driver.f5.pool)

    def test_member(self):
        DriverTest(self,
                   self.driver.member,
                   FakeModel("member-01"),
                   self.driver.f5.member)

    def test_health_monitor(self):
        DriverTest(self,
                   self.driver.health_monitor,
                   FakeModel("hm-01"),
                   self.driver.f5.healthmonitor)

    def test_l7policy(self):
        DriverTest(self,
                   self.driver.l7policy,
                   FakeModel("l7policy-01"),
                   self.driver.f5.l7policy)

    def test_l7rule(self):
        DriverTest(self,
                   self.driver.l7rule,
                   FakeModel("l7rule-01"),
                   self.driver.f5.l7rule)



# Copyright 2013 New Dream Network, LLC (DreamHost)
# Copyright 2015 Rackspace
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import socket

import mock
from neutron.agent.linux import external_process
from neutron_lib import constants
from neutron_lib import exceptions

from neutron_lbaas.drivers.haproxy import namespace_driver
from neutron_lbaas.services.loadbalancer import data_models
from neutron_lbaas.tests import base
from oslo_utils import fileutils


class TestHaproxyNSDriver(base.BaseTestCase):

    def setUp(self):
        super(TestHaproxyNSDriver, self).setUp()

        conf = mock.Mock()
        conf.haproxy.loadbalancer_state_path = '/the/path'
        conf.interface_driver = 'intdriver'
        conf.haproxy.user_group = 'test_group'
        conf.haproxy.send_gratuitous_arp = 3
        self.conf = conf
        self.rpc_mock = mock.Mock()
        self.ensure_tree = mock.patch.object(fileutils, 'ensure_tree').start()
        self._process_monitor = mock.Mock()
        with mock.patch(
                'neutron_lib.utils.runtime.load_class_by_alias_or_classname'):
            self.driver = namespace_driver.HaproxyNSDriver(
                conf,
                self.rpc_mock,
                self._process_monitor
            )
        self.vif_driver = mock.Mock()
        self.driver.vif_driver = self.vif_driver
        self._build_mock_data_models()

    def _build_mock_data_models(self):
        host_route = data_models.HostRoute(destination='0.0.0.0/0',
                                           nexthop='192.0.0.1')
        subnet = data_models.Subnet(cidr='10.0.0.1/24',
                                    gateway_ip='10.0.0.2',
                                    host_routes=[host_route])
        fixed_ip = data_models.IPAllocation(ip_address='10.0.0.1')
        setattr(fixed_ip, 'subnet', subnet)
        port = data_models.Port(id='port1', network_id='network1',
                                mac_address='12-34-56-78-9A-BC',
                                fixed_ips=[fixed_ip])
        self.lb = data_models.LoadBalancer(id='lb1', listeners=[],
                                           vip_port=port,
                                           vip_address='10.0.0.1')

    def test_get_name(self):
        self.assertEqual(namespace_driver.DRIVER_NAME, self.driver.get_name())

    @mock.patch('neutron.agent.linux.ip_lib.IPWrapper')
    @mock.patch('os.makedirs')
    @mock.patch('os.path.dirname')
    @mock.patch('os.path.isdir')
    @mock.patch('shutil.rmtree')
    def test_undeploy_instance(self, mock_shutil, mock_isdir, mock_dirname,
                               mock_makedirs, mock_ip_wrap):
        self.driver._get_state_file_path = mock.Mock(return_value='/path')
        self.driver._unplug = mock.Mock()
        mock_dirname.return_value = '/path/' + self.lb.id
        mock_isdir.return_value = False
        self.driver.undeploy_instance(self.lb.id)
        calls = [mock.call(self.lb.id, 'pid'), mock.call(self.lb.id, '')]
        self.driver._get_state_file_path.has_calls(calls)
        self.assertFalse(self.driver._unplug.called)
        self.assertFalse(mock_ip_wrap.called)
        mock_isdir.assert_called_once_with('/path/' + self.lb.id)
        self.assertFalse(mock_shutil.called)

        self.driver.deployed_loadbalancers[self.lb.id] = self.lb
        mock_isdir.return_value = True
        mock_isdir.reset_mock()
        mock_ns = mock_ip_wrap.return_value
        mock_ns.get_devices.return_value = [collections.namedtuple(
            'Device', ['name'])(name='test_device')]
        self.driver.undeploy_instance(self.lb.id, cleanup_namespace=True,
                                      delete_namespace=True)
        ns = namespace_driver.get_ns_name(self.lb.id)
        calls = [mock.call(self.lb.id, 'pid'), mock.call(self.lb.id, '')]
        self.driver._get_state_file_path.has_calls(calls)
        self.driver._unplug.assert_called_once_with(ns, self.lb.vip_port)
        ip_wrap_calls = [mock.call(namespace=ns), mock.call(namespace=ns)]
        mock_ip_wrap.has_calls(ip_wrap_calls)
        mock_ns.get_devices.assert_called_once_with(exclude_loopback=True)
        self.vif_driver.unplug.assert_called_once_with('test_device',
                                                       namespace=ns)
        mock_shutil.assert_called_once_with('/path/' + self.lb.id)
        mock_ns.garbage_collect_namespace.assert_called_once_with()

    @mock.patch('os.makedirs')
    @mock.patch('os.path.dirname')
    def test_undeploy_instance_unregister_usage(self, mock_dirname,
                                                mock_makedirs):
        self.driver._get_state_file_path = mock.Mock(return_value='/path')
        self.driver._unplug = mock.Mock()
        mock_dirname.return_value = '/path/' + self.lb.id
        with mock.patch.object(self._process_monitor,
                               'unregister') as mock_unregister:
            self.driver.undeploy_instance(self.lb.id)
            mock_unregister.assert_called_once_with(
                    uuid=self.lb.id, service_name='lbaas-ns-haproxy')

    @mock.patch('os.path.exists')
    @mock.patch('os.listdir')
    def test_remove_orphans(self, list_dir, exists):
        lb_ids = [self.lb.id]
        exists.return_value = False
        self.driver.remove_orphans(lb_ids)
        exists.assert_called_once_with(self.driver.state_path)
        self.assertFalse(list_dir.called)

        exists.reset_mock()
        exists.return_value = True
        list_dir.return_value = [self.lb.id, 'lb2']
        self.driver.exists = mock.Mock()
        self.driver.undeploy_instance = mock.Mock()
        self.driver.remove_orphans(lb_ids)
        exists.assert_called_once_with(self.driver.state_path)
        list_dir.assert_called_once_with(self.driver.state_path)
        self.driver.exists.assert_called_once_with('lb2')
        self.driver.undeploy_instance.assert_called_once_with(
            'lb2', cleanup_namespace=True)

    def test_get_stats(self):
        # Shamelessly stolen from v1 namespace driver tests.
        raw_stats = ('# pxname,svname,qcur,qmax,scur,smax,slim,stot,bin,bout,'
                     'dreq,dresp,ereq,econ,eresp,wretr,wredis,status,weight,'
                     'act,bck,chkfail,chkdown,lastchg,downtime,qlimit,pid,iid,'
                     'sid,throttle,lbtot,tracked,type,rate,rate_lim,rate_max,'
                     'check_status,check_code,check_duration,hrsp_1xx,'
                     'hrsp_2xx,hrsp_3xx,hrsp_4xx,hrsp_5xx,hrsp_other,hanafail,'
                     'req_rate,req_rate_max,req_tot,cli_abrt,srv_abrt,\n'
                     '8e271901-69ed-403e-a59b-f53cf77ef208,BACKEND,1,2,3,4,0,'
                     '10,7764,2365,0,0,,0,0,0,0,UP,1,1,0,,0,103780,0,,1,2,0,,0'
                     ',,1,0,,0,,,,0,0,0,0,0,0,,,,,0,0,\n\n'
                     'a557019b-dc07-4688-9af4-f5cf02bb6d4b,'
                     '32a6c2a3-420a-44c3-955d-86bd2fc6871e,0,0,0,1,,7,1120,'
                     '224,,0,,0,0,0,0,UP,1,1,0,0,1,2623,303,,1,2,1,,7,,2,0,,'
                     '1,L7OK,200,98,0,7,0,0,0,0,0,,,,0,0,\n'
                     'a557019b-dc07-4688-9af4-f5cf02bb6d4b,'
                     'd9aea044-8867-4e80-9875-16fb808fa0f9,0,0,0,2,,12,0,0,,'
                     '0,,0,0,8,4,DOWN,1,1,0,9,2,308,675,,1,2,2,,4,,2,0,,2,'
                     'L4CON,,2999,0,0,0,0,0,0,0,,,,0,0,\n')
        raw_stats_empty = ('# pxname,svname,qcur,qmax,scur,smax,slim,stot,bin,'
                           'bout,dreq,dresp,ereq,econ,eresp,wretr,wredis,'
                           'status,weight,act,bck,chkfail,chkdown,lastchg,'
                           'downtime,qlimit,pid,iid,sid,throttle,lbtot,'
                           'tracked,type,rate,rate_lim,rate_max,check_status,'
                           'check_code,check_duration,hrsp_1xx,hrsp_2xx,'
                           'hrsp_3xx,hrsp_4xx,hrsp_5xx,hrsp_other,hanafail,'
                           'req_rate,req_rate_max,req_tot,cli_abrt,srv_abrt,'
                           '\n')
        with mock.patch.object(self.driver, '_get_state_file_path') as gsp, \
                mock.patch('socket.socket') as mocket, \
                mock.patch('os.path.exists') as path_exists, \
                mock.patch.object(data_models.LoadBalancer, 'from_dict') \
                as lb_from_dict, \
                mock.patch.object(self.driver, '_is_active') as is_active:
            gsp.side_effect = lambda x, y, z: '/pool/' + y
            path_exists.return_value = True
            mocket.return_value = mocket
            mocket.recv.return_value = raw_stats.encode('utf-8')
            is_active.return_value = True

            exp_stats = {'connection_errors': '0',
                         'active_connections': '3',
                         'current_sessions': '3',
                         'bytes_in': '7764',
                         'max_connections': '4',
                         'max_sessions': '4',
                         'bytes_out': '2365',
                         'response_errors': '0',
                         'total_sessions': '10',
                         'total_connections': '10',
                         'members': {
                             '32a6c2a3-420a-44c3-955d-86bd2fc6871e': {
                                 'operating_status': 'ONLINE',
                                 'health': 'L7OK',
                                 'failed_checks': '0'
                             },
                             'd9aea044-8867-4e80-9875-16fb808fa0f9': {
                                 'operating_status': 'OFFLINE',
                                 'health': 'L4CON',
                                 'failed_checks': '9'
                             }
                         }
                         }
            stats = self.driver.get_stats(self.lb.id)
            self.assertEqual(exp_stats, stats)

            mocket.recv.return_value = raw_stats_empty.encode('utf-8')
            self.assertEqual({'members': {}},
                             self.driver.get_stats(self.lb.id))

            path_exists.return_value = False
            is_active.return_value = True
            listener = data_models.Listener(
                provisioning_status=constants.PENDING_CREATE,
                admin_state_up=True)
            self.lb.listeners.append(listener)
            lb_from_dict.return_value = \
                data_models.LoadBalancer.from_dict(self.lb)

            self.lb.listeners.append(listener)
            self.assertEqual({}, self.driver.get_stats(self.lb.id))

            path_exists.return_value = False
            mocket.reset_mock()
            is_active.return_value = False
            self.assertEqual({}, self.driver.get_stats(self.lb.id))
            self.assertFalse(mocket.called)

    def test_is_active(self):
        # test no listeners
        ret_val = self.driver._is_active(self.lb)
        self.assertFalse(ret_val)

        # test bad VIP status
        listener = data_models.Listener(
            provisioning_status=constants.PENDING_CREATE,
            admin_state_up=True)
        self.lb.listeners.append(listener)
        self.lb.vip_port.status = constants.DOWN
        ret_val = self.driver._is_active(self.lb)
        self.assertFalse(ret_val)
        self.lb.vip_port.status = constants.PENDING_CREATE
        self.lb.vip_port.admin_state_up = False
        ret_val = self.driver._is_active(self.lb)
        self.assertFalse(ret_val)

        # test bad LB status
        self.lb.vip_port.admin_state_up = True
        self.lb.operating_status = 'OFFLINE'
        ret_val = self.driver._is_active(self.lb)
        self.assertFalse(ret_val)

        # test everything good
        self.lb.operating_status = 'ONLINE'
        ret_val = self.driver._is_active(self.lb)
        self.assertTrue(ret_val)

    def test_deploy_instance(self):
        self.driver.deployable = mock.Mock(return_value=False)
        self.driver.exists = mock.Mock(return_value=True)
        self.driver.update = mock.Mock()
        self.driver.create = mock.Mock()

        def reset():
            self.driver.deployable.reset_mock()
            self.driver.exists.reset_mock()
            self.driver.update.reset_mock()
            self.driver.create.reset_mock()

        deployed = self.driver.deploy_instance(self.lb)
        self.assertFalse(deployed)
        self.assertFalse(self.driver.exists.called)
        self.assertFalse(self.driver.create.called)
        self.assertFalse(self.driver.update.called)

        reset()
        self.driver.deployable.return_value = True
        deployed = self.driver.deploy_instance(self.lb)
        self.assertTrue(deployed)
        self.driver.exists.assert_called_once_with(self.lb.id)
        self.driver.update.assert_called_once_with(self.lb)
        self.assertFalse(self.driver.create.called)

        reset()
        self.driver.exists.return_value = False
        deployed = self.driver.deploy_instance(self.lb)
        self.assertTrue(deployed)
        self.driver.exists.assert_called_once_with(self.lb.id)
        self.driver.create.assert_called_once_with(self.lb)
        self.assertFalse(self.driver.update.called)

    def test_update(self):
        self.driver._get_state_file_path = mock.Mock(return_value='/path')
        self.driver._spawn = mock.Mock()
        with mock.patch('six.moves.builtins.open') as m_open:
            file_mock = mock.MagicMock()
            m_open.return_value = file_mock
            file_mock.__enter__.return_value = file_mock
            file_mock.__iter__.return_value = iter(['123'])
            self.driver.update(self.lb)
            self.driver._spawn.assert_called_once_with(self.lb,
                                                       ['-sf', '123'])

    @mock.patch('socket.socket')
    @mock.patch('os.path.exists')
    @mock.patch('neutron.agent.linux.ip_lib.IPWrapper')
    def test_exists(self, ip_wrap, exists, mocket):
        socket_path = '/path/haproxy_stats.sock'
        mock_ns = ip_wrap.return_value
        mock_socket = mocket.return_value
        self.driver._get_state_file_path = mock.Mock(return_value=socket_path)
        mock_ns.netns.exists.return_value = False
        exists.return_value = False

        def reset():
            ip_wrap.reset_mock()
            self.driver._get_state_file_path.reset_mock()
            mock_ns.reset_mock()
            exists.reset_mock()
            mocket.reset_mock()
            mock_socket.reset_mock()

        ret_exists = self.driver.exists(self.lb.id)
        ip_wrap.assert_called_once_with()
        self.driver._get_state_file_path.assert_called_once_with(
            self.lb.id, 'haproxy_stats.sock', False)
        mock_ns.netns.exists.assert_called_once_with(
            namespace_driver.get_ns_name(self.lb.id))
        self.assertFalse(exists.called)
        self.assertFalse(mocket.called)
        self.assertFalse(mock_socket.connect.called)
        self.assertFalse(ret_exists)

        reset()
        mock_ns.netns.exists.return_value = True
        exists.return_value = False
        ret_exists = self.driver.exists(self.lb.id)
        ip_wrap.assert_called_once_with()
        self.driver._get_state_file_path.assert_called_once_with(
            self.lb.id, 'haproxy_stats.sock', False)
        mock_ns.netns.exists.assert_called_once_with(
            namespace_driver.get_ns_name(self.lb.id))
        exists.assert_called_once_with(socket_path)
        self.assertFalse(mocket.called)
        self.assertFalse(mock_socket.connect.called)
        self.assertFalse(ret_exists)

        reset()
        mock_ns.netns.exists.return_value = True
        exists.return_value = True
        ret_exists = self.driver.exists(self.lb.id)
        ip_wrap.assert_called_once_with()
        self.driver._get_state_file_path.assert_called_once_with(
            self.lb.id, 'haproxy_stats.sock', False)
        mock_ns.netns.exists.assert_called_once_with(
            namespace_driver.get_ns_name(self.lb.id))
        exists.assert_called_once_with(socket_path)
        mocket.assert_called_once_with(socket.AF_UNIX, socket.SOCK_STREAM)
        mock_socket.connect.assert_called_once_with(socket_path)
        self.assertTrue(ret_exists)

    def test_create(self):
        self.driver._plug = mock.Mock()
        self.driver._spawn = mock.Mock()
        self.driver.create(self.lb)
        self.driver._plug.assert_called_once_with(
            namespace_driver.get_ns_name(self.lb.id),
            self.lb.vip_port, self.lb.vip_address)
        self.driver._spawn.assert_called_once_with(self.lb)

    def test_deployable(self):
        # test None
        ret_val = self.driver.deployable(None)
        self.assertFalse(ret_val)

        # test no listeners
        ret_val = self.driver.deployable(self.lb)
        self.assertFalse(ret_val)

        # test no acceptable listeners
        listener = data_models.Listener(
            provisioning_status=constants.PENDING_DELETE,
            admin_state_up=True)
        self.lb.listeners.append(listener)
        ret_val = self.driver.deployable(self.lb)
        self.assertFalse(ret_val)
        listener.provisioning_status = constants.PENDING_CREATE
        listener.admin_state_up = False
        ret_val = self.driver.deployable(self.lb)
        self.assertFalse(ret_val)

        # test bad lb status
        listener.admin_state_up = True
        self.lb.provisioning_status = constants.PENDING_DELETE
        self.lb.admin_state_up = True
        ret_val = self.driver.deployable(self.lb)
        self.assertFalse(ret_val)
        self.lb.provisioning_status = constants.PENDING_UPDATE
        self.lb.admin_state_up = False
        ret_val = self.driver.deployable(self.lb)
        self.assertFalse(ret_val)

        # test everything good
        self.lb.admin_state_up = True
        ret_val = self.driver.deployable(self.lb)
        self.assertTrue(ret_val)

    @mock.patch('oslo_utils.fileutils.ensure_tree')
    def test_get_state_file_path(self, ensure_tree):
        path = self.driver._get_state_file_path(self.lb.id, 'conf',
                                                ensure_state_dir=False)
        self.assertEqual('/the/path/v2/lb1/conf', path)
        self.assertFalse(ensure_tree.called)
        path = self.driver._get_state_file_path(self.lb.id, 'conf')
        self.assertEqual('/the/path/v2/lb1/conf', path)
        self.assertTrue(ensure_tree.called)

    @mock.patch('neutron.agent.linux.ip_lib.device_exists')
    @mock.patch('neutron.agent.linux.ip_lib.IPWrapper')
    def test_plug(self, ip_wrap, device_exists):
        device_exists.return_value = True
        interface_name = 'tap-d4nc3'
        self.vif_driver.get_device_name.return_value = interface_name
        self.assertRaises(exceptions.PreexistingDeviceFailure,
                          self.driver._plug, 'ns1', self.lb.vip_port,
                          self.lb.vip_address, reuse_existing=False)
        device_exists.assert_called_once_with(interface_name,
                                              namespace='ns1')
        self.rpc_mock.plug_vip_port.assert_called_once_with(
            self.lb.vip_port.id)

        device_exists.reset_mock()
        self.rpc_mock.plug_vip_port.reset_mock()
        mock_ns = ip_wrap.return_value
        self.driver._plug('ns1', self.lb.vip_port, self.lb.vip_address)
        self.rpc_mock.plug_vip_port.assert_called_once_with(
            self.lb.vip_port.id)
        device_exists.assert_called_once_with(interface_name,
                                              namespace='ns1')
        self.assertFalse(self.vif_driver.plug.called)
        expected_cidrs = ['10.0.0.1/24']
        self.vif_driver.init_l3.assert_called_once_with(
            interface_name, expected_cidrs, namespace='ns1')
        calls = [mock.call(['route', 'add', 'default', 'gw', '192.0.0.1'],
                           check_exit_code=False),
                 mock.call(['arping', '-U', '-I', interface_name,
                            '-c', 3, '10.0.0.1'],
                           check_exit_code=False)]
        mock_ns.netns.execute.has_calls(calls)
        self.assertEqual(2, mock_ns.netns.execute.call_count)

    def test_unplug(self):
        interface_name = 'tap-d4nc3'
        self.vif_driver.get_device_name.return_value = interface_name
        self.driver._unplug('ns1', self.lb.vip_port)
        self.rpc_mock.unplug_vip_port.assert_called_once_with(
            self.lb.vip_port.id)
        self.vif_driver.get_device_name.assert_called_once_with(
            self.lb.vip_port)
        self.vif_driver.unplug.assert_called_once_with(interface_name,
                                                       namespace='ns1')

    @mock.patch('oslo_utils.fileutils.ensure_tree')
    @mock.patch('neutron_lbaas.drivers.haproxy.jinja_cfg.save_config')
    @mock.patch('neutron.agent.linux.ip_lib.IPWrapper')
    def test_spawn(self, ip_wrap, jinja_save, ensure_tree):
        mock_ns = ip_wrap.return_value
        self.driver._spawn(self.lb)
        conf_dir = self.driver.state_path + '/' + self.lb.id + '/%s'
        jinja_save.assert_called_once_with(
            conf_dir % 'haproxy.conf',
            self.lb,
            conf_dir % 'haproxy_stats.sock',
            'test_group',
            conf_dir % '')
        ip_wrap.assert_called_once_with(
            namespace=namespace_driver.get_ns_name(self.lb.id))
        mock_ns.netns.execute.assert_called_once_with(
            ['haproxy', '-f', conf_dir % 'haproxy.conf', '-p',
             conf_dir % 'haproxy.pid'], addl_env=None, run_as_root=True)
        self.assertIn(self.lb.id, self.driver.deployed_loadbalancers)
        self.assertEqual(self.lb,
                         self.driver.deployed_loadbalancers[self.lb.id])

    @mock.patch('oslo_utils.fileutils.ensure_tree')
    @mock.patch('neutron_lbaas.drivers.haproxy.jinja_cfg.save_config')
    def test_spawn_enable_usage(self, jinja_save, ensure_tree):
        with mock.patch.object(external_process.ProcessManager,
                               'enable') as mock_enable:
            self.driver._spawn(self.lb)
            mock_enable.assert_called_once_with()

    @mock.patch('oslo_utils.fileutils.ensure_tree')
    @mock.patch('neutron_lbaas.drivers.haproxy.jinja_cfg.save_config')
    def test_spawn_reload_cfg_usage(self, jinja_save, ensure_tree):
        with mock.patch.object(external_process.ProcessManager, 'active',
                               return_value=True):
            with mock.patch.object(external_process.ProcessManager,
                                   'reload_cfg') as mock_reload_cfg:
                extra_cmd_args = ['-sf', '123']
                self.driver._spawn(self.lb, extra_cmd_args=extra_cmd_args)
                mock_reload_cfg.assert_called_once_with()


class BaseTestManager(base.BaseTestCase):

    def setUp(self):
        super(BaseTestManager, self).setUp()
        self.driver = mock.Mock()
        self.lb_manager = namespace_driver.LoadBalancerManager(self.driver)
        self.listener_manager = namespace_driver.ListenerManager(self.driver)
        self.pool_manager = namespace_driver.PoolManager(self.driver)
        self.member_manager = namespace_driver.MemberManager(self.driver)
        self.hm_manager = namespace_driver.HealthMonitorManager(self.driver)
        self.refresh = self.driver.loadbalancer.refresh


class BaseTestLoadBalancerManager(BaseTestManager):

    def setUp(self):
        super(BaseTestLoadBalancerManager, self).setUp()
        self.in_lb = data_models.LoadBalancer(id='lb1', listeners=[])


class TestLoadBalancerManager(BaseTestLoadBalancerManager):

    @mock.patch.object(data_models.LoadBalancer, 'from_dict')
    def test_refresh(self, lb_from_dict):
        rpc_return = {'id': self.in_lb.id}
        self.driver.plugin_rpc.get_loadbalancer.return_value = rpc_return
        from_dict_return = data_models.LoadBalancer(id=self.in_lb.id)
        lb_from_dict.return_value = from_dict_return
        self.driver.deploy_instance.return_value = True
        self.driver.exists.return_value = True
        self.lb_manager.refresh(self.in_lb)
        self.driver.plugin_rpc.get_loadbalancer.assert_called_once_with(
            self.in_lb.id)
        lb_from_dict.assert_called_once_with(rpc_return)
        self.driver.deploy_instance.assert_called_once_with(from_dict_return)
        self.assertFalse(self.driver.exists.called)
        self.assertFalse(self.driver.undeploy_instance.called)

        self.driver.reset_mock()
        lb_from_dict.reset_mock()
        self.driver.deploy_instance.return_value = False
        self.driver.exists.return_value = False
        self.lb_manager.refresh(self.in_lb)
        self.driver.plugin_rpc.get_loadbalancer.assert_called_once_with(
            self.in_lb.id)
        lb_from_dict.assert_called_once_with(rpc_return)
        self.driver.deploy_instance.assert_called_once_with(from_dict_return)
        self.driver.exists.assert_called_once_with(self.in_lb.id)
        self.assertFalse(self.driver.undeploy_instance.called)

        self.driver.reset_mock()
        lb_from_dict.reset_mock()
        self.driver.deploy_instance.return_value = False
        self.driver.exists.return_value = True
        self.lb_manager.refresh(self.in_lb)
        self.driver.plugin_rpc.get_loadbalancer.assert_called_once_with(
            self.in_lb.id)
        lb_from_dict.assert_called_once_with(rpc_return)
        self.driver.deploy_instance.assert_called_once_with(from_dict_return)
        self.driver.exists.assert_called_once_with(from_dict_return.id)
        self.driver.undeploy_instance.assert_called_once_with(self.in_lb.id)

    def test_delete(self):
        self.driver.exists.return_value = False
        self.lb_manager.delete(self.in_lb)
        self.driver.exists.assert_called_once_with(self.in_lb.id)
        self.assertFalse(self.driver.undeploy_instance.called)

        self.driver.reset_mock()
        self.driver.exists.return_value = True
        self.lb_manager.delete(self.in_lb)
        self.driver.exists.assert_called_once_with(self.in_lb.id)
        self.driver.undeploy_instance.assert_called_once_with(
            self.in_lb.id, delete_namespace=True)

    def test_create(self):
        self.lb_manager.refresh = mock.Mock()
        self.lb_manager.create(self.in_lb)
        self.assertFalse(self.lb_manager.refresh.called)

        self.lb_manager.refresh.reset_mock()
        self.in_lb.listeners.append(data_models.Listener(id='listener1'))
        self.lb_manager.create(self.in_lb)
        self.lb_manager.refresh.assert_called_once_with(self.in_lb)

    def test_get_stats(self):
        self.driver.get_stats.return_value = {'members': {}}
        self.lb_manager.get_stats(self.in_lb.id)
        self.driver.get_stats.assert_called_once_with(self.in_lb.id)

    def test_update(self):
        old_lb = data_models.LoadBalancer(id='lb0')
        self.lb_manager.refresh = mock.Mock()
        self.lb_manager.update(old_lb, self.in_lb)
        self.lb_manager.refresh.assert_called_once_with(self.in_lb)


class BaseTestListenerManager(BaseTestLoadBalancerManager):

    def setUp(self):
        super(BaseTestListenerManager, self).setUp()
        self.in_listener = data_models.Listener(id='listener1')
        self.listener2 = data_models.Listener(id='listener2')
        self.in_listener.loadbalancer = self.in_lb
        self.listener2.loadbalancer = self.in_lb
        self.in_lb.listeners = [self.in_listener, self.listener2]
        self.refresh = self.driver.loadbalancer.refresh


class TestListenerManager(BaseTestListenerManager):

    def setUp(self):
        super(TestListenerManager, self).setUp()
        self.in_listener = data_models.Listener(id='listener1')
        self.listener2 = data_models.Listener(id='listener2')
        self.in_lb.listeners = [self.in_listener, self.listener2]
        self.in_listener.loadbalancer = self.in_lb
        self.listener2.loadbalancer = self.in_lb

    def test_remove_listener(self):
        self.listener_manager._remove_listener(self.in_lb, self.in_listener.id)
        self.assertEqual(1, len(self.in_lb.listeners))
        self.assertEqual(self.listener2.id, self.in_lb.listeners[0].id)

    def test_update(self):
        old_listener = data_models.Listener(id='listener1', name='bleh')
        self.listener_manager.update(old_listener, self.in_listener)
        self.refresh.assert_called_once_with(self.in_lb)

    def test_create(self):
        self.listener_manager.create(self.in_listener)
        self.refresh.assert_called_once_with(self.in_lb)

    def test_delete(self):
        self.listener_manager.delete(self.in_listener)
        self.refresh.assert_called_once_with(self.in_lb)
        self.assertFalse(self.driver.undeploy_instance.called)

        self.refresh.reset_mock()
        self.driver.reset_mock()
        self.listener_manager.delete(self.listener2)
        self.assertFalse(self.refresh.called)
        self.driver.undeploy_instance.assert_called_once_with(
            self.in_lb.id, delete_namespace=True)


class BaseTestPoolManager(BaseTestListenerManager):

    def setUp(self):
        super(BaseTestPoolManager, self).setUp()
        self.in_pool = data_models.Pool(id='pool1')
        self.in_listener.default_pool = self.in_pool
        self.in_pool.loadbalancer = self.in_lb
        self.in_pool.listeners = [self.in_listener]
        self.in_lb.pools = [self.in_pool]


class TestPoolManager(BaseTestPoolManager):

    def test_update(self):
        old_pool = data_models.Pool(id=self.in_pool.id, name='bleh')
        self.pool_manager.update(old_pool, self.in_pool)
        self.refresh.assert_called_once_with(self.in_lb)

    def test_create(self):
        self.pool_manager.create(self.in_pool)
        self.refresh.assert_called_once_with(self.in_lb)

    def test_delete(self):
        self.pool_manager.delete(self.in_pool)
        self.assertIsNone(self.in_listener.default_pool)
        self.refresh.assert_called_once_with(self.in_lb)


class BaseTestMemberManager(BaseTestPoolManager):

    def setUp(self):
        super(BaseTestMemberManager, self).setUp()
        self.in_member = data_models.Member(id='member1')
        self.member2 = data_models.Member(id='member2')
        self.in_pool.members = [self.in_member, self.member2]
        self.in_member.pool = self.in_pool
        self.member2.pool = self.in_pool


class TestMemberManager(BaseTestMemberManager):

    def test_remove_member(self):
        self.member_manager._remove_member(self.in_pool, self.in_member.id)
        self.assertEqual(1, len(self.in_pool.members))
        self.assertEqual(self.member2.id, self.in_pool.members[0].id)

    def test_update(self):
        old_member = data_models.Member(id=self.in_member.id,
                                        address='0.0.0.0')
        self.member_manager.update(old_member, self.in_member)
        self.refresh.assert_called_once_with(self.in_lb)

    def test_create(self):
        self.member_manager.create(self.in_member)
        self.refresh.assert_called_once_with(self.in_lb)

    def test_delete(self):
        self.member_manager.delete(self.in_member)
        self.refresh.assert_called_once_with(self.in_lb)


class BaseTestHealthMonitorManager(BaseTestPoolManager):

    def setUp(self):
        super(BaseTestHealthMonitorManager, self).setUp()
        self.in_hm = data_models.HealthMonitor(id='hm1')
        self.in_pool.healthmonitor = self.in_hm
        self.in_hm.pool = self.in_pool


class TestHealthMonitorManager(BaseTestHealthMonitorManager):

    def test_update(self):
        old_hm = data_models.HealthMonitor(id=self.in_hm.id, timeout=2)
        self.hm_manager.update(old_hm, self.in_hm)
        self.refresh.assert_called_once_with(self.in_lb)

    def test_create(self):
        self.hm_manager.create(self.in_hm)
        self.refresh.assert_called_once_with(self.in_lb)

    def test_delete(self):
        self.hm_manager.delete(self.in_hm)
        self.assertIsNone(self.in_pool.healthmonitor)
        self.refresh.assert_called_once_with(self.in_lb)


class TestNamespaceDriverModule(base.BaseTestCase):

    def test_get_ns_name(self):
        ns_name = namespace_driver.get_ns_name('woohoo')
        self.assertEqual(namespace_driver.NS_PREFIX + 'woohoo', ns_name)




# Copyright 2015 Citrix Systems, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import abc

from neutron_lib import constants
from neutron_lib import context as ncontext
from oslo_config import cfg
from oslo_log import log as logging
from oslo_service import service

from neutron_lbaas._i18n import _
from neutron_lbaas.drivers import driver_base
from neutron_lbaas.drivers import driver_mixins
from neutron_lbaas.drivers.netscaler import ncc_client

DEFAULT_PERIODIC_TASK_INTERVAL = "2"
DEFAULT_STATUS_COLLECTION = "True"
DEFAULT_PAGE_SIZE = "300"
DEFAULT_IS_SYNCRONOUS = "True"

PROV = "provisioning_status"
NETSCALER = "netscaler"

LOG = logging.getLogger(__name__)

NETSCALER_CC_OPTS = [
    cfg.StrOpt(
        'netscaler_ncc_uri',
        help=_('The URL to reach the NetScaler Control Center Server.'),
    ),
    cfg.StrOpt(
        'netscaler_ncc_username',
        help=_('Username to login to the NetScaler Control Center Server.'),
    ),
    cfg.StrOpt(
        'netscaler_ncc_password',
        secret=True,
        help=_('Password to login to the NetScaler Control Center Server.'),
    ),
    cfg.StrOpt(
        'periodic_task_interval',
        default=DEFAULT_PERIODIC_TASK_INTERVAL,
        help=_('Setting for periodic task collection interval from'
               'NetScaler Control Center Server..'),
    ),
    cfg.StrOpt(
        'is_synchronous',
        default=DEFAULT_IS_SYNCRONOUS,
        help=_('Setting for option to enable synchronous operations'
               'NetScaler Control Center Server.'),
    ),
    cfg.StrOpt(
        'netscaler_ncc_cleanup_mode',
        help=_(
            'Setting to enable/disable cleanup mode for NetScaler Control '
            'Center Server'),
    ),
    cfg.StrOpt(
        'netscaler_status_collection',
        default=DEFAULT_STATUS_COLLECTION + "," + DEFAULT_PAGE_SIZE,
        help=_('Setting for member status collection from'
               'NetScaler Control Center Server.'),
    )
]


if not hasattr(cfg.CONF, "netscaler_driver"):
    cfg.CONF.register_opts(NETSCALER_CC_OPTS, 'netscaler_driver')


LBS_RESOURCE = 'loadbalancers'
LB_RESOURCE = 'loadbalancer'
LISTENERS_RESOURCE = 'listeners'
LISTENER_RESOURCE = 'listener'
POOLS_RESOURCE = 'pools'
POOL_RESOURCE = 'pool'
MEMBERS_RESOURCE = 'members'
MEMBER_RESOURCE = 'member'
MONITORS_RESOURCE = 'healthmonitors'
MONITOR_RESOURCE = 'healthmonitor'
STATS_RESOURCE = 'stats'
PROV_SEGMT_ID = 'provider:segmentation_id'
PROV_NET_TYPE = 'provider:network_type'
DRIVER_NAME = 'netscaler_driver'
RESOURCE_PREFIX = 'v2.0/lbaas'
STATUS_PREFIX = 'oca/v2'
MEMBER_STATUS = 'memberstatus'
PAGE = 'page'
SIZE = 'size'


PROVISIONING_STATUS_TRACKER = []


class NetScalerLoadBalancerDriverV2(driver_base.LoadBalancerBaseDriver):

    def __init__(self, plugin):
        super(NetScalerLoadBalancerDriverV2, self).__init__(plugin)

        self.driver_conf = cfg.CONF.netscaler_driver
        self.admin_ctx = ncontext.get_admin_context()
        self._init_client()
        self._init_managers()
        self._init_status_collection()

    def _init_client(self):
        ncc_uri = self.driver_conf.netscaler_ncc_uri
        ncc_username = self.driver_conf.netscaler_ncc_username
        ncc_password = self.driver_conf.netscaler_ncc_password
        ncc_cleanup_mode = cfg.CONF.netscaler_driver.netscaler_ncc_cleanup_mode
        self.client = ncc_client.NSClient(ncc_uri,
                                          ncc_username,
                                          ncc_password,
                                          ncc_cleanup_mode)

    def _init_managers(self):
        self.load_balancer = NetScalerLoadBalancerManager(self)
        self.listener = NetScalerListenerManager(self)
        self.pool = NetScalerPoolManager(self)
        self.member = NetScalerMemberManager(self)
        self.health_monitor = NetScalerHealthMonitorManager(self)

    def _init_status_collection(self):
        self.status_conf = self.driver_conf.netscaler_status_collection
        self.periodic_task_interval = self.driver_conf.periodic_task_interval
        status_conf = self.driver_conf.netscaler_status_collection
        (is_status_collection,
            pagesize_status_collection) = status_conf.split(",")
        self.is_status_collection = True
        if is_status_collection.lower() == "false":
            self.is_status_collection = False
        self.pagesize_status_collection = pagesize_status_collection

        self._init_pending_status_tracker()

        NetScalerStatusService(self).start()

    def _init_pending_status_tracker(self):
        # Initialize PROVISIONING_STATUS_TRACKER for loadbalancers in
        # pending state
        db_lbs = self.plugin.db.get_loadbalancers(
            self.admin_ctx)
        for db_lb in db_lbs:
            if ((db_lb.id not in PROVISIONING_STATUS_TRACKER) and
                (db_lb.provider.provider_name == NETSCALER) and
                    (db_lb.provisioning_status.startswith("PENDING_"))):
                PROVISIONING_STATUS_TRACKER.append(db_lb.id)

    def collect_provision_status(self):

        msg = ("Collecting status ", self.periodic_task_interval)
        LOG.debug(msg)
        self._update_loadbalancers_provision_status()

    def _update_loadbalancers_provision_status(self):
        for lb_id in PROVISIONING_STATUS_TRACKER:
            lb_statuses = self._get_loadbalancer_statuses(lb_id)
            if lb_statuses:
                self._update_status_tree_in_db(
                    lb_id, lb_statuses["lb_statuses"])

    def _get_loadbalancer_statuses(self, lb_id):
        """Retrieve listener status from Control Center."""
        resource_path = "%s/%s/%s/statuses" % (RESOURCE_PREFIX,
                                               LBS_RESOURCE,
                                               lb_id)
        try:
            statuses = self.client.retrieve_resource(
                "GLOBAL", resource_path)[1]['dict']
        except ncc_client.NCCException as e:
            if e.is_not_found_exception():
                return {"lb_statuses": None}
            else:
                return None
        statuses = statuses["statuses"]
        return {"lb_statuses": statuses}

    def _update_status_tree_in_db(self, lb_id, loadbalancer_statuses):
        track_loadbalancer = {"track": False}
        db_lb = self.plugin.db.get_loadbalancer(self.admin_ctx,
                                                lb_id)

        if (not loadbalancer_statuses and
                db_lb.provisioning_status == constants.PENDING_DELETE):
            try:
                self.load_balancer.successful_completion(
                    self.admin_ctx, db_lb, delete=True)
            except Exception:
                LOG.error("error with successful completion")
            PROVISIONING_STATUS_TRACKER.remove(lb_id)
            return
        else:
            status_lb = loadbalancer_statuses["loadbalancer"]

        status_listeners = status_lb["listeners"]
        for db_listener in db_lb.listeners:
            db_listener.loadbalancer = db_lb
            status_listener = (self.
                               _update_entity_status_in_db(track_loadbalancer,
                                                           db_listener,
                                                           status_listeners,
                                                           self.listener))
            if not status_listener:
                continue

            db_pool = db_listener.default_pool

            if not db_pool:
                continue
            db_pool.listener = db_listener

            status_pools = status_listener['pools']
            status_pool = self._update_entity_status_in_db(track_loadbalancer,
                                                           db_pool,
                                                           status_pools,
                                                           self.pool)

            db_members = db_pool.members
            if not status_pool:
                continue
            status_members = status_pool['members']

            for db_member in db_members:
                db_member.pool = db_pool
                self._update_entity_status_in_db(track_loadbalancer,
                                                 db_member,
                                                 status_members,
                                                 self.member)

            db_hm = db_pool.healthmonitor
            if db_hm:
                db_hm.pool = db_pool
                status_hm = status_pool['healthmonitor']
                self._update_entity_status_in_db(track_loadbalancer,
                                                 db_hm,
                                                 [status_hm],
                                                 self.health_monitor)

        if not track_loadbalancer['track']:
            self._update_entity_status_in_db(
                track_loadbalancer, db_lb, status_lb, self.load_balancer)
            if not track_loadbalancer['track']:
                PROVISIONING_STATUS_TRACKER.remove(lb_id)

    def _update_entity_status_in_db(self, track_loadbalancer,
                                    db_entity,
                                    status_entities,
                                    entity_manager):
        if isinstance(status_entities, list):
            entity_status = self._get_entity_status(
                db_entity.id, status_entities)
        else:
            entity_status = status_entities

        self._check_and_update_entity_status_in_db(
            track_loadbalancer, db_entity, entity_status, entity_manager)
        return entity_status

    def _get_entity_status(self, entity_id, entities_status):
        for entity_status in entities_status:
            if entity_status and entity_status['id'] == entity_id:
                return entity_status
        return None

    def _check_and_update_entity_status_in_db(self, track_loadbalancer,
                                              db_entity,
                                              entity_status, entity_manager):

        if not db_entity.provisioning_status.startswith("PENDING_"):
            # no operation is attempted on this entity
            return
        if entity_status:
            if entity_status[PROV].startswith("PENDING_"):
                # an entity is not finished provisioning. Continue to track
                track_loadbalancer['track'] = True
                return

            if entity_status[PROV] == constants.ERROR:
                # Marked for failed completion
                try:
                    entity_manager.failed_completion(
                        self.admin_ctx, db_entity)
                except Exception:
                    LOG.error("error with failed completion")
                return

        if db_entity.provisioning_status == constants.PENDING_DELETE:
            # entity is under deletion
            # if entity is missing in lb status tree it should to be
            # deleted
            if entity_status:
                msg = ('Invalid status set for delete of %s in statuses',
                       db_entity.id)
                LOG.error(msg)
                return
            try:
                entity_manager.successful_completion(
                    self.admin_ctx, db_entity, delete=True)
            except Exception:
                LOG.error("error with successful completion")
            return

        if entity_status[PROV] != constants.ACTIVE:
            msg = ('Invalid prov status for %s, should be ACTIVE '
                   "for CREATE and UPDATE",
                   db_entity.id)
            LOG.error(msg)
            return
        try:
            entity_manager.successful_completion(
                self.admin_ctx, db_entity)
        except Exception:
            LOG.error("error with successful completion")

        return


class NetScalerCommonManager(driver_mixins.BaseManagerMixin):

    def __init__(self, driver):
        super(NetScalerCommonManager, self).__init__(driver)
        self.payload_preparer = PayloadPreparer()
        self.client = self.driver.client

        self.is_synchronous = self.driver.driver_conf.is_synchronous
        if self.is_synchronous.lower() == "false":
            self.is_synchronous = False
        else:
            self.is_synchronous = True

    def create(self, context, obj):
        LOG.debug("%s, create %s", self.__class__.__name__, obj.id)
        try:
            self.create_entity(context, obj)
            if self.is_synchronous:
                self.successful_completion(context, obj)
            else:
                self.track_provision_status(obj)
        except Exception:
            self.failed_completion(context, obj)
            raise

    def update(self, context, old_obj, obj):
        LOG.debug("%s, update %s", self.__class__.__name__, old_obj.id)
        try:
            self.update_entity(context, old_obj, obj)
            if self.is_synchronous:
                self.successful_completion(context, obj)
            else:
                self.track_provision_status(obj)
        except Exception:
            self.failed_completion(context, obj)
            raise

    def delete(self, context, obj):
        LOG.debug("%s, delete %s", self.__class__.__name__, obj.id)
        try:
            self.delete_entity(context, obj)
            if self.is_synchronous:
                self.successful_completion(context, obj, delete=True)
            else:
                self.track_provision_status(obj)
        except Exception:
            self.failed_completion(context, obj)
            raise

    def track_provision_status(self, obj):
        for lb in self._get_loadbalancers(obj):
            if lb.id not in PROVISIONING_STATUS_TRACKER:
                PROVISIONING_STATUS_TRACKER.append(lb.id)

    def _get_loadbalancers(self, obj):
        lbs = []
        lbs.append(obj.root_loadbalancer)
        return lbs

    @abc.abstractmethod
    def create_entity(self, context, obj):
        pass

    @abc.abstractmethod
    def update_entity(self, context, old_obj, obj):
        pass

    @abc.abstractmethod
    def delete_entity(self, context, obj):
        pass


class NetScalerLoadBalancerManager(NetScalerCommonManager,
                                   driver_base.BaseLoadBalancerManager):

    def __init__(self, driver):
        driver_base.BaseLoadBalancerManager.__init__(self, driver)
        NetScalerCommonManager.__init__(self, driver)

    def refresh(self, context, lb_obj):
        # This is intended to trigger the backend to check and repair
        # the state of this load balancer and all of its dependent objects
        LOG.debug("LB refresh %s", lb_obj.id)

    def stats(self, context, lb_obj):
        pass

    def create_entity(self, context, lb_obj):
        ncc_lb = self.payload_preparer.prepare_lb_for_creation(lb_obj)
        vip_subnet_id = lb_obj.vip_subnet_id
        network_info = self.payload_preparer.\
            get_network_info(context, self.driver.plugin, vip_subnet_id)
        ncc_lb.update(network_info)
        msg = _("NetScaler driver lb creation: %s") % repr(ncc_lb)
        LOG.debug(msg)
        resource_path = "%s/%s" % (RESOURCE_PREFIX, LBS_RESOURCE)
        self.client.create_resource(context.tenant_id, resource_path,
                                    LB_RESOURCE, ncc_lb)

    def update_entity(self, context, old_lb_obj, lb_obj):
        update_lb = self.payload_preparer.prepare_lb_for_update(lb_obj)
        resource_path = "%s/%s/%s" % (RESOURCE_PREFIX, LBS_RESOURCE, lb_obj.id)
        msg = (_("NetScaler driver lb_obj %(lb_obj_id)s update: %(lb_obj)s") %
               {"lb_obj_id": old_lb_obj.id, "lb_obj": repr(lb_obj)})
        LOG.debug(msg)
        self.client.update_resource(context.tenant_id, resource_path,
                                    LB_RESOURCE, update_lb)

    def delete_entity(self, context, lb_obj):
        """Delete a loadbalancer on a NetScaler device."""
        resource_path = "%s/%s/%s" % (RESOURCE_PREFIX, LBS_RESOURCE, lb_obj.id)
        msg = _("NetScaler driver lb_obj removal: %s") % lb_obj.id
        LOG.debug(msg)
        self.client.remove_resource(context.tenant_id, resource_path)


class NetScalerListenerManager(NetScalerCommonManager,
                               driver_base.BaseListenerManager):

    def __init__(self, driver):
        driver_base.BaseListenerManager.__init__(self, driver)
        NetScalerCommonManager.__init__(self, driver)

    def stats(self, context, listener):
        # returning dummy status now
        LOG.debug(
            "Tenant id %s , Listener stats %s", context.tenant_id, listener.id)
        return {
            "bytes_in": 0,
            "bytes_out": 0,
            "active_connections": 0,
            "total_connections": 0
        }

    def create_entity(self, context, listener):
        """Listener is created with loadbalancer """
        ncc_listener = self.payload_preparer.prepare_listener_for_creation(
            listener)
        msg = _("NetScaler driver listener creation: %s") % repr(ncc_listener)
        LOG.debug(msg)
        resource_path = "%s/%s" % (RESOURCE_PREFIX, LISTENERS_RESOURCE)
        self.client.create_resource(context.tenant_id, resource_path,
                                    LISTENER_RESOURCE, ncc_listener)

    def update_entity(self, context, old_listener, listener):
        update_listener = self.payload_preparer.prepare_listener_for_update(
            listener)
        resource_path = "%s/%s/%s" % (RESOURCE_PREFIX, LISTENERS_RESOURCE,
                                      listener.id)
        msg = (_("NetScaler driver listener %(listener_id)s "
                 "update: %(listener_obj)s") %
               {"listener_id": old_listener.id,
                "listener_obj": repr(listener)})
        LOG.debug(msg)
        self.client.update_resource(context.tenant_id, resource_path,
                                    LISTENER_RESOURCE, update_listener)

    def delete_entity(self, context, listener):
        """Delete a listener on a NetScaler device."""
        resource_path = "%s/%s/%s" % (RESOURCE_PREFIX, LISTENERS_RESOURCE,
                                      listener.id)
        msg = _("NetScaler driver listener removal: %s") % listener.id
        LOG.debug(msg)
        self.client.remove_resource(context.tenant_id, resource_path)


class NetScalerPoolManager(NetScalerCommonManager,
                           driver_base.BasePoolManager):

    def __init__(self, driver):
        driver_base.BasePoolManager.__init__(self, driver)
        NetScalerCommonManager.__init__(self, driver)

    def create_entity(self, context, pool):
        ncc_pool = self.payload_preparer.prepare_pool_for_creation(
            pool)
        msg = _("NetScaler driver pool creation: %s") % repr(ncc_pool)
        LOG.debug(msg)
        resource_path = "%s/%s" % (RESOURCE_PREFIX, POOLS_RESOURCE)
        self.client.create_resource(context.tenant_id, resource_path,
                                    POOL_RESOURCE, ncc_pool)

    def update_entity(self, context, old_pool, pool):
        update_pool = self.payload_preparer.prepare_pool_for_update(pool)
        resource_path = "%s/%s/%s" % (RESOURCE_PREFIX, POOLS_RESOURCE,
                                      pool.id)
        msg = (_("NetScaler driver pool %(pool_id)s update: %(pool_obj)s") %
               {"pool_id": old_pool.id, "pool_obj": repr(pool)})
        LOG.debug(msg)
        self.client.update_resource(context.tenant_id, resource_path,
                                    POOL_RESOURCE, update_pool)

    def delete_entity(self, context, pool):
        """Delete a pool on a NetScaler device."""
        resource_path = "%s/%s/%s" % (RESOURCE_PREFIX, POOLS_RESOURCE,
                                      pool.id)
        msg = _("NetScaler driver pool removal: %s") % pool.id
        LOG.debug(msg)
        self.client.remove_resource(context.tenant_id, resource_path)


class NetScalerMemberManager(NetScalerCommonManager,
                             driver_base.BaseMemberManager):

    def __init__(self, driver):
        driver_base.BaseMemberManager.__init__(self, driver)
        NetScalerCommonManager.__init__(self, driver)

    def create_entity(self, context, member):

        ncc_member = self.payload_preparer.prepare_member_for_creation(member)
        subnet_id = member.subnet_id
        network_info = (self.payload_preparer.
                        get_network_info(context, self.driver.plugin,
                                         subnet_id))
        ncc_member.update(network_info)
        msg = _("NetScaler driver member creation: %s") % repr(ncc_member)
        LOG.debug(msg)
        parent_pool_id = member.pool.id
        resource_path = "%s/%s/%s/%s" % (RESOURCE_PREFIX, POOLS_RESOURCE,
                                         parent_pool_id, MEMBERS_RESOURCE)
        self.client.create_resource(context.tenant_id, resource_path,
                                    MEMBER_RESOURCE, ncc_member)

    def update_entity(self, context, old_member, member):
        parent_pool_id = member.pool.id
        update_member = self.payload_preparer.prepare_member_for_update(member)
        resource_path = "%s/%s/%s/%s/%s" % (RESOURCE_PREFIX,
                                            POOLS_RESOURCE,
                                            parent_pool_id,
                                            MEMBERS_RESOURCE,
                                            member.id)
        msg = (_("NetScaler driver member %(member_id)s "
                 "update: %(member_obj)s") %
               {"member_id": old_member.id, "member_obj": repr(member)})
        LOG.debug(msg)
        self.client.update_resource(context.tenant_id, resource_path,
                                    MEMBER_RESOURCE, update_member)

    def delete_entity(self, context, member):
        """Delete a member on a NetScaler device."""
        parent_pool_id = member.pool.id
        resource_path = "%s/%s/%s/%s/%s" % (RESOURCE_PREFIX,
                                            POOLS_RESOURCE,
                                            parent_pool_id,
                                            MEMBERS_RESOURCE,
                                            member.id)
        msg = _("NetScaler driver member removal: %s") % member.id
        LOG.debug(msg)
        self.client.remove_resource(context.tenant_id, resource_path)


class NetScalerHealthMonitorManager(NetScalerCommonManager,
                                    driver_base.BaseHealthMonitorManager):

    def __init__(self, driver):
        driver_base.BaseHealthMonitorManager.__init__(self, driver)
        NetScalerCommonManager.__init__(self, driver)

    def create_entity(self, context, hm):
        ncc_hm = self.payload_preparer.prepare_healthmonitor_for_creation(hm)
        msg = _("NetScaler driver healthmonitor creation: %s") % repr(ncc_hm)
        LOG.debug(msg)
        resource_path = "%s/%s" % (RESOURCE_PREFIX, MONITORS_RESOURCE)
        self.client.create_resource(context.tenant_id, resource_path,
                                    MONITOR_RESOURCE, ncc_hm)

    def update_entity(self, context, old_healthmonitor, hm):
        update_hm = self.payload_preparer.prepare_healthmonitor_for_update(hm)
        resource_path = "%s/%s/%s" % (RESOURCE_PREFIX, MONITORS_RESOURCE,
                                      hm.id)
        msg = (_("NetScaler driver healthmonitor %(healthmonitor_id)s "
                 "update: %(healthmonitor_obj)s") %
               {"healthmonitor_id": old_healthmonitor.id,
                "healthmonitor_obj": repr(hm)})
        LOG.debug(msg)
        self.client.update_resource(context.tenant_id, resource_path,
                                    MONITOR_RESOURCE, update_hm)

    def delete_entity(self, context, hm):
        """Delete a healthmonitor on a NetScaler device."""
        resource_path = "%s/%s/%s" % (RESOURCE_PREFIX, MONITORS_RESOURCE,
                                      hm.id)
        msg = _("NetScaler driver healthmonitor removal: %s") % hm.id
        LOG.debug(msg)
        self.client.remove_resource(context.tenant_id, resource_path)


class PayloadPreparer(object):

    def prepare_lb_for_creation(self, lb):
        creation_attrs = {
            'id': lb.id,
            'tenant_id': lb.tenant_id,
            'vip_address': lb.vip_address,
            'vip_subnet_id': lb.vip_subnet_id,
        }
        update_attrs = self.prepare_lb_for_update(lb)
        creation_attrs.update(update_attrs)

        return creation_attrs

    def prepare_lb_for_update(self, lb):
        return {
            'name': lb.name,
            'description': lb.description,
            'admin_state_up': lb.admin_state_up,
        }

    def prepare_listener_for_creation(self, listener):
        creation_attrs = {
            'id': listener.id,
            'tenant_id': listener.tenant_id,
            'protocol': listener.protocol,
            'protocol_port': listener.protocol_port,
            'loadbalancer_id': listener.loadbalancer_id
        }
        update_attrs = self.prepare_listener_for_update(listener)
        creation_attrs.update(update_attrs)
        return creation_attrs

    def prepare_listener_for_update(self, listener):
        sni_container_ids = self.prepare_sni_container_ids(listener)
        listener_dict = {
            'name': listener.name,
            'description': listener.description,
            'sni_container_ids': sni_container_ids,
            'default_tls_container_id': listener.default_tls_container_id,
            'connection_limit': listener.connection_limit,
            'admin_state_up': listener.admin_state_up
        }
        return listener_dict

    def prepare_pool_for_creation(self, pool):
        create_attrs = {
            'id': pool.id,
            'tenant_id': pool.tenant_id,
            'listener_id': pool.listener.id,
            'protocol': pool.protocol,
        }
        update_attrs = self.prepare_pool_for_update(pool)
        create_attrs.update(update_attrs)
        return create_attrs

    def prepare_pool_for_update(self, pool):
        update_attrs = {
            'name': pool.name,
            'description': pool.description,
            'lb_algorithm': pool.lb_algorithm,
            'admin_state_up': pool.admin_state_up
        }
        if pool.session_persistence:
            peristence = pool.session_persistence
            peristence_payload = self.prepare_sessionpersistence(peristence)
            update_attrs['session_persistence'] = peristence_payload
        return update_attrs

    def prepare_sessionpersistence(self, persistence):
        return {
            'type': persistence.type,
            'cookie_name': persistence.cookie_name
        }

    def prepare_members_for_pool(self, members):
        members_attrs = []
        for member in members:
            member_attrs = self.prepare_member_for_creation(member)
            members_attrs.append(member_attrs)
        return members_attrs

    def prepare_member_for_creation(self, member):
        creation_attrs = {
            'id': member.id,
            'tenant_id': member.tenant_id,
            'address': member.address,
            'protocol_port': member.protocol_port,
            'subnet_id': member.subnet_id
        }
        update_attrs = self.prepare_member_for_update(member)
        creation_attrs.update(update_attrs)
        return creation_attrs

    def prepare_member_for_update(self, member):
        return {
            'weight': member.weight,
            'admin_state_up': member.admin_state_up,
        }

    def prepare_healthmonitor_for_creation(self, health_monitor):
        creation_attrs = {
            'id': health_monitor.id,
            'tenant_id': health_monitor.tenant_id,
            'pool_id': health_monitor.pool.id,
            'type': health_monitor.type,
        }
        update_attrs = self.prepare_healthmonitor_for_update(health_monitor)
        creation_attrs.update(update_attrs)
        return creation_attrs

    def prepare_healthmonitor_for_update(self, health_monitor):
        ncc_hm = {
            'delay': health_monitor.delay,
            'timeout': health_monitor.timeout,
            'max_retries': health_monitor.max_retries,
            'admin_state_up': health_monitor.admin_state_up,
        }
        if health_monitor.type in ['HTTP', 'HTTPS']:
            ncc_hm['http_method'] = health_monitor.http_method
            ncc_hm['url_path'] = health_monitor.url_path
            ncc_hm['expected_codes'] = health_monitor.expected_codes
        return ncc_hm

    def get_network_info(self, context, plugin, subnet_id):
        network_info = {}
        subnet = plugin.db._core_plugin.get_subnet(context, subnet_id)
        network_id = subnet['network_id']
        network = plugin.db._core_plugin.get_network(context, network_id)
        network_info['network_id'] = network_id
        network_info['subnet_id'] = subnet_id
        if PROV_NET_TYPE in network:
            network_info['network_type'] = network[PROV_NET_TYPE]
        if PROV_SEGMT_ID in network:
            network_info['segmentation_id'] = network[PROV_SEGMT_ID]
        return network_info

    def prepare_sni_container_ids(self, listener):
        sni_container_ids = []
        for sni_container in listener.sni_containers:
            sni_container_ids.append(sni_container.tls_container_id)
        return sni_container_ids


class NetScalerStatusService(service.Service):

    def __init__(self, driver):
        super(NetScalerStatusService, self).__init__()
        self.driver = driver

    def start(self):
        super(NetScalerStatusService, self).start()
        self.tg.add_timer(
            int(self.driver.periodic_task_interval),
            self.driver.collect_provision_status,
            None
        )



# Copyright 2015, 2016 Rackspace US Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tempest.common import utils
from tempest import config
from tempest.lib.common.utils import data_utils
from tempest.lib import decorators
from tempest.lib import exceptions as ex

from neutron_lbaas.tests.tempest.v2.api import base

CONF = config.CONF


class MemberTestJSON(base.BaseTestCase):

    """
    Test the following operations in Neutron-LBaaS API using the
    REST client for members:

        list members of a pool
        create a member of a Pool
        update a pool member
        delete a member
    """

    @classmethod
    def resource_setup(cls):
        super(MemberTestJSON, cls).resource_setup()
        if not utils.is_extension_enabled("lbaasv2", "network"):
            msg = "lbaas extension not enabled."
            raise cls.skipException(msg)
        network_name = data_utils.rand_name('network-')
        cls.network = cls.create_network(network_name)
        cls.subnet = cls.create_subnet(cls.network)
        cls.tenant_id = cls.subnet.get('tenant_id')
        cls.subnet_id = cls.subnet.get('id')
        cls.load_balancer = cls._create_active_load_balancer(
            tenant_id=cls.tenant_id,
            vip_subnet_id=cls.subnet.get('id'))
        cls.load_balancer_id = cls.load_balancer.get("id")
        cls.listener = cls._create_listener(
            loadbalancer_id=cls.load_balancer.get('id'),
            protocol=cls.listener_protocol, protocol_port=80)
        cls.listener_id = cls.listener.get('id')
        cls.pool = cls._create_pool(protocol=cls.pool_protocol,
                                    tenant_id=cls.tenant_id,
                                    lb_algorithm='ROUND_ROBIN',
                                    listener_id=cls.listener_id)
        cls.pool_id = cls.pool.get('id')

    @classmethod
    def resource_cleanup(cls):
        super(MemberTestJSON, cls).resource_cleanup()

    def test_list_empty_members(self):
        """Test that pool members are empty."""
        members = self.members_client.list_members(self.pool_id)
        self.assertEmpty(members,
                         msg='Initial pool was supposed to be empty')

    @decorators.attr(type='smoke')
    def test_list_3_members(self):
        """Test that we can list members. """
        member_ips_exp = set([u"127.0.0.0", u"127.0.0.1", u"127.0.0.2"])
        for ip in member_ips_exp:
            member_opts = self.build_member_opts()
            member_opts["address"] = ip
            member = self._create_member(self.pool_id, **member_opts)
            self.addCleanup(self._delete_member, self.pool_id, member['id'])
        members = self.members_client.list_members(self.pool_id)
        self.assertEqual(3, len(members))
        for member in members:
            self.assertEqual(member["tenant_id"], self.tenant_id)
            self.assertEqual(member["protocol_port"], 80)
            self.assertEqual(member["subnet_id"], self.subnet_id)
        found_member_ips = set([m["address"] for m in members])
        self.assertEqual(found_member_ips, member_ips_exp)

    @decorators.attr(type='smoke')
    def test_add_member(self):
        """Test that we can add a single member."""
        expect_empty_members = self.members_client.list_members(self.pool_id)
        self.assertEmpty(expect_empty_members)
        member_opts = self.build_member_opts()
        member = self._create_member(self.pool_id, **member_opts)
        member_id = member.get("id")
        self.addCleanup(self._delete_member, self.pool_id, member_id)
        self.assertEqual(member_opts["address"], member["address"])
        self.assertEqual(self.tenant_id, member["tenant_id"])
        self.assertEqual(80, member["protocol_port"])
        self.assertEqual(self.subnet_id, member["subnet_id"])
        # Should have default values for admin_state_up and weight
        self.assertTrue(member["admin_state_up"])
        self.assertEqual(1, member["weight"])

    @decorators.attr(type='smoke')
    def test_get_member(self):
        """Test that we can fetch a member by id."""
        member_opts = self.build_member_opts()
        member_id = self._create_member(self.pool_id,
                                        **member_opts)["id"]
        self.addCleanup(self._delete_member, self.pool_id, member_id)
        member = self.members_client.get_member(self.pool_id, member_id)
        self.assertEqual(member_id, member["id"])
        self.assertEqual(member_opts["address"], member["address"])
        self.assertEqual(member_opts["tenant_id"], member["tenant_id"])
        self.assertEqual(member_opts["protocol_port"], member["protocol_port"])
        self.assertEqual(member_opts["subnet_id"], member["subnet_id"])

    def test_create_member_missing_required_field_tenant_id(self):
        """Test if a non_admin user can create a member with tenant_id
        missing
        """
        member_opts = {}
        member_opts['address'] = "127.0.0.1"
        member_opts['protocol_port'] = 80
        member_opts['subnet_id'] = self.subnet_id
        member = self._create_member(self.pool_id, **member_opts)
        self.addCleanup(self._delete_member, self.pool_id, member['id'])

    @decorators.attr(type='negative')
    def test_create_member_missing_required_field_address(self):
        """Test create a member with missing field address"""
        member_opts = {}
        member_opts['protocol_port'] = 80
        member_opts['subnet_id'] = self.subnet_id
        self.assertRaises(ex.BadRequest, self._create_member,
                          self.pool_id, **member_opts)

    @decorators.attr(type='negative')
    def test_create_member_missing_required_field_protocol_port(self):
        """Test create a member with missing field protocol_port"""
        member_opts = {}
        member_opts['address'] = "127.0.0.1"
        member_opts['subnet_id'] = self.subnet_id
        self.assertRaises(ex.BadRequest, self._create_member,
                          self.pool_id, **member_opts)

    @decorators.skip_because(bug="1468457")
    @decorators.attr(type='negative')
    def test_create_member_missing_required_field_subnet_id(self):
        """Test create a member with missing field subnet_id """
        member_opts = {}
        member_opts['protocol_port'] = 80
        member_opts['address'] = "127.0.0.1"
        self.assertRaises(ex.BadRequest, self._create_member,
                          self.pool_id, **member_opts)

    @decorators.attr(type='negative')
    def test_raises_BadRequest_when_missing_attrs_during_member_create(self):
        """Test failure on missing attributes on member create."""
        member_opts = {}
        self.assertRaises(ex.BadRequest, self._create_member,
                          self.pool_id, **member_opts)

    @decorators.skip_because(bug="1468457")
    @decorators.attr(type='negative')
    def test_create_member_invalid_tenant_id(self):
        """Test create member with invalid tenant_id"""
        member_opts = {}
        member_opts['address'] = "127.0.0.1"
        member_opts['protocol_port'] = 80
        member_opts['subnet_id'] = self.subnet_id
        member_opts['tenant_id'] = "$232!$pw"
        self.assertRaises(ex.BadRequest, self._create_member,
                          self.pool_id, **member_opts)

    @decorators.attr(type='negative')
    def test_create_member_invalid_address(self):
        """Test create member with invalid address"""
        member_opts = {}
        member_opts['address'] = "127$%<ki"
        member_opts['protocol_port'] = 80
        member_opts['subnet_id'] = self.subnet_id
        self.assertRaises(ex.BadRequest, self._create_member,
                          self.pool_id, **member_opts)

    @decorators.attr(type='negative')
    def test_create_member_invalid_protocol_port(self):
        """Test create member with invalid protocol_port"""
        member_opts = {}
        member_opts['address'] = "127.0.0.1"
        member_opts['protocol_port'] = 8090000
        member_opts['subnet_id'] = self.subnet_id
        self.assertRaises(ex.BadRequest, self._create_member,
                          self.pool_id, **member_opts)

    @decorators.attr(type='negative')
    def test_create_member_invalid_subnet_id(self):
        """Test create member with invalid subnet_id"""
        member_opts = {}
        member_opts['address'] = "127.0.0.1"
        member_opts['protocol_port'] = 80
        member_opts['subnet_id'] = "45k%^"
        self.assertRaises(ex.BadRequest, self._create_member,
                          self.pool_id, **member_opts)

    @decorators.attr(type='negative')
    def test_create_member_invalid_admin_state_up(self):
        """Test create member with invalid admin_state_up"""
        member_opts = {}
        member_opts['address'] = "127.0.0.1"
        member_opts['protocol_port'] = 80
        member_opts['subnet_id'] = self.subnet_id
        member_opts['admin_state_up'] = "$232!$pw"
        self.assertRaises(ex.BadRequest, self._create_member,
                          self.pool_id, **member_opts)

    @decorators.attr(type='negative')
    def test_create_member_invalid_weight(self):
        """Test create member with invalid weight"""
        member_opts = {}
        member_opts['address'] = "127.0.0.1"
        member_opts['protocol_port'] = 80
        member_opts['subnet_id'] = self.subnet_id
        member_opts['weight'] = "$232!$pw"
        self.assertRaises(ex.BadRequest, self._create_member,
                          self.pool_id, **member_opts)

    @decorators.skip_because(bug="1468457")  # Octavia does a floor()
    @decorators.attr(type='negative')
    def test_create_member_nonint_weight(self):
        """Test create member with nonint weight"""
        member_opts = {}
        member_opts['address'] = "127.0.0.1"
        member_opts['protocol_port'] = 80
        member_opts['subnet_id'] = self.subnet_id
        member_opts['weight'] = 0.1
        self.assertRaises(ex.BadRequest, self._create_member,
                          self.pool_id, **member_opts)

    @decorators.skip_because(bug="1468457")
    @decorators.attr(type='negative')
    def test_create_member_empty_tenant_id(self):
        """Test create member with an empty tenant_id"""
        member_opts = {}
        member_opts['address'] = "127.0.0.1"
        member_opts['protocol_port'] = 80
        member_opts['subnet_id'] = self.subnet_id
        member_opts['tenant_id'] = ""
        self.assertRaises(ex.BadRequest, self._create_member,
                          self.pool_id, **member_opts)

    @decorators.attr(type='negative')
    def test_create_member_empty_address(self):
        """Test create member with an empty address"""
        member_opts = {}
        member_opts['address'] = ""
        member_opts['protocol_port'] = 80
        member_opts['subnet_id'] = self.subnet_id
        self.assertRaises(ex.BadRequest, self._create_member,
                          self.pool_id, **member_opts)

    @decorators.attr(type='negative')
    def test_create_member_empty_protocol_port(self):
        """Test create member with an empty protocol_port"""
        member_opts = {}
        member_opts['address'] = "127.0.0.1"
        member_opts['protocol_port'] = ""
        member_opts['subnet_id'] = self.subnet_id
        self.assertRaises(ex.BadRequest, self._create_member,
                          self.pool_id, **member_opts)

    @decorators.attr(type='negative')
    def test_create_member_empty_subnet_id(self):
        """Test create member with empty subnet_id"""
        member_opts = {}
        member_opts['address'] = "127.0.0.1"
        member_opts['protocol_port'] = 80
        member_opts['subnet_id'] = ""
        self.assertRaises(ex.BadRequest, self._create_member,
                          self.pool_id, **member_opts)

    @decorators.attr(type='negative')
    def test_create_member_empty_admin_state_up(self):
        """Test create member with an empty admin_state_up"""
        member_opts = {}
        member_opts['address'] = "127.0.0.1"
        member_opts['protocol_port'] = 80
        member_opts['subnet_id'] = self.subnet_id
        member_opts['admin_state_up'] = ""
        self.assertRaises(ex.BadRequest, self._create_member,
                          self.pool_id, **member_opts)

    @decorators.attr(type='negative')
    def test_create_member_empty_weight(self):
        """Test create member with an empty weight"""
        member_opts = {}
        member_opts['address'] = "127.0.0.1"
        member_opts['protocol_port'] = 80
        member_opts['subnet_id'] = self.subnet_id
        member_opts['weight'] = ""
        self.assertRaises(ex.BadRequest, self._create_member,
                          self.pool_id, **member_opts)

    @decorators.attr(type='smoke')
    def test_delete_member(self):
        """Test that we can delete a member by id."""
        member_opts = self.build_member_opts()
        member_id = self._create_member(self.pool_id,
                                        **member_opts)["id"]
        members = self.members_client.list_members(self.pool_id)
        self.assertEqual(1, len(members))
        self._delete_member(self.pool_id, member_id)
        members = self.members_client.list_members(self.pool_id)
        self.assertEmpty(members)

    @decorators.attr(type='smoke')
    def test_update_member(self):
        """Test that we can update a member."""
        member_opts = self.build_member_opts()
        member = self._create_member(self.pool_id,
                                     **member_opts)
        member_id = member["id"]
        self.addCleanup(self._delete_member, self.pool_id, member['id'])
        # Make sure the defaults are correct
        self.assertTrue(member["admin_state_up"])
        self.assertEqual(1, member["weight"])
        # Lets overwrite the defaults
        member_opts = {"weight": 10, "admin_state_up": False}
        self._update_member(self.pool_id, member_id,
                            **member_opts)
        member = self.members_client.get_member(self.pool_id, member_id)
        # And make sure they stick
        self.assertFalse(member["admin_state_up"])
        self.assertEqual(10, member["weight"])

    def test_update_member_missing_admin_state_up(self):
        """Test that we can update a member with missing admin_state_up."""
        member_opts = self.build_member_opts()
        member = self._create_member(self.pool_id, **member_opts)
        member_id = member["id"]
        self.addCleanup(self._delete_member, self.pool_id, member_id)
        self.assertTrue(member["admin_state_up"])
        self.assertEqual(1, member["weight"])
        member_opts = {"weight": 10}
        self._update_member(self.pool_id, member_id, **member_opts)
        member = self.members_client.get_member(self.pool_id, member_id)
        self.assertTrue(member["admin_state_up"])
        self.assertEqual(10, member["weight"])

    def test_update_member_missing_weight(self):
        """Test that we can update a member with missing weight."""
        member_opts = self.build_member_opts()
        member = self._create_member(self.pool_id,
                                     **member_opts)
        member_id = member["id"]
        self.addCleanup(self._delete_member, self.pool_id, member_id)
        self.assertTrue(member["admin_state_up"])
        self.assertEqual(1, member["weight"])
        member_opts = {"admin_state_up": False}
        self._update_member(self.pool_id, member_id, **member_opts)
        member = self.members_client.get_member(self.pool_id, member_id)
        self.assertFalse(member["admin_state_up"])
        self.assertEqual(1, member["weight"])

    @decorators.attr(type='negative')
    def test_update_member_invalid_admin_state_up(self):
        """Test that we can update a member with empty admin_state_up."""
        member_opts = self.build_member_opts()
        member = self._create_member(self.pool_id,
                                     **member_opts)
        member_id = member["id"]
        self.addCleanup(self._delete_member, self.pool_id, member_id)
        self.assertTrue(member["admin_state_up"])
        self.assertEqual(1, member["weight"])
        member_opts = {"weight": 10, "admin_state_up": "%^67"}
        self.assertRaises(ex.BadRequest, self._update_member,
                          self.pool_id, member_id, **member_opts)

    @decorators.attr(type='negative')
    def test_update_member_invalid_weight(self):
        """Test that we can update a member with an empty weight."""
        member_opts = self.build_member_opts()
        member = self._create_member(self.pool_id,
                                     **member_opts)
        member_id = member["id"]
        self.addCleanup(self._delete_member, self.pool_id, member_id)
        self.assertTrue(member["admin_state_up"])
        self.assertEqual(1, member["weight"])
        member_opts = {"admin_state_up": False, "weight": "*^$df"}
        self.assertRaises(ex.BadRequest, self._update_member,
                          self.pool_id, member_id, **member_opts)

    @decorators.attr(type='negative')
    def test_update_member_empty_admin_state_up(self):
        """Test that we can update a member with empty admin_state_up."""
        member_opts = self.build_member_opts()
        member = self._create_member(self.pool_id,
                                     **member_opts)
        member_id = member["id"]
        self.addCleanup(self._delete_member, self.pool_id, member_id)
        self.assertTrue(member["admin_state_up"])
        self.assertEqual(1, member["weight"])
        member_opts = {"weight": 10, "admin_state_up": ""}
        self.assertRaises(ex.BadRequest, self._update_member,
                          self.pool_id, member_id, **member_opts)

    @decorators.attr(type='negative')
    def test_update_member_empty_weight(self):
        """Test that we can update a member with an empty weight."""
        member_opts = self.build_member_opts()
        member = self._create_member(self.pool_id,
                                     **member_opts)
        member_id = member["id"]
        self.addCleanup(self._delete_member, self.pool_id, member_id)
        self.assertTrue(member["admin_state_up"])
        self.assertEqual(1, member["weight"])
        member_opts = {"admin_state_up": False, "weight": ""}
        self.assertRaises(ex.BadRequest, self._update_member,
                          self.pool_id, member_id, **member_opts)

    @decorators.attr(type='negative')
    def test_raises_immutable_when_updating_immutable_attrs_on_member(self):
        """Test failure on immutable attribute on member create."""
        member_opts = self.build_member_opts()
        member_id = self._create_member(self.pool_id,
                                        **member_opts)["id"]
        self.addCleanup(self._delete_member, self.pool_id, member_id)
        member_opts = {"address": "127.0.0.69"}
        # The following code actually raises a 400 instead of a 422 as expected
        # Will need to consult with blogan as to what to fix
        self.assertRaises(ex.BadRequest, self._update_member,
                          self.pool_id, member_id, **member_opts)

    @decorators.attr(type='negative')
    def test_raises_exception_on_invalid_attr_on_create(self):
        """Test failure on invalid attribute on member create."""
        member_opts = self.build_member_opts()
        member_opts["invalid_op"] = "should_break_request"
        self.assertRaises(ex.BadRequest, self._create_member,
                          self.pool_id, **member_opts)

    @decorators.attr(type='negative')
    def test_raises_exception_on_invalid_attr_on_update(self):
        """Test failure on invalid attribute on member update."""
        member_opts = self.build_member_opts()
        member = self._create_member(self.pool_id, **member_opts)
        member_id = member["id"]
        self.addCleanup(self._delete_member, self.pool_id, member_id)
        member_opts["invalid_op"] = "watch_this_break"
        self.assertRaises(ex.BadRequest, self._update_member,
                          self.pool_id, member_id, **member_opts)

    @classmethod
    def build_member_opts(cls, **kw):
        """Build out default member dictionary """
        opts = {"address": kw.get("address", "127.0.0.1"),
                "tenant_id": kw.get("tenant_id", cls.tenant_id),
                "protocol_port": kw.get("protocol_port", 80),
                "subnet_id": kw.get("subnet_id", cls.subnet_id)}
        return opts



# Copyright 2014 Brocade Communications Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# Pattabi Ayyasami (pattabi), Brocade Communication Systems, Inc.
#
import sys

import mock
from neutron_lib import context

from neutron_lbaas.services.loadbalancer import data_models
from neutron_lbaas.tests.unit.db.loadbalancer import test_db_loadbalancerv2
with mock.patch.dict(sys.modules, {'brocade_neutron_lbaas': mock.Mock()}):
    from neutron_lbaas.drivers.brocade import driver_v2 as driver


class FakeModel(object):
    def __init__(self, id):
        self.id = id

    def attached_to_loadbalancer(self):
        return True


class ManagerTest(object):
    def __init__(self, parent, manager, model):
        self.parent = parent
        self.manager = manager
        self.model = model

        self.create(model)
        self.update(model, model)
        self.delete(model)

    def create(self, model):
        self.manager.create(self.parent.context, model)

    def update(self, old_model, model):
        self.manager.update(self.parent.context, old_model, model)

    def delete(self, model):
        self.manager.delete(self.parent.context, model)


class LoadBalancerManagerTest(ManagerTest):
    def __init__(self, parent, manager, model):
        super(LoadBalancerManagerTest, self).__init__(parent, manager, model)

        self.refresh(model)
        self.stats(model)

    def refresh(self, model):
        self.manager.refresh(self.parent.context, model)
        self.parent.driver.device_driver.refresh \
            .assert_called_once_with(model)

    def stats(self, model):
        self.manager.stats(self.parent.context, model)
        self.parent.driver.device_driver.stats.assert_called_once_with(model)


class TestBrocadeLoadBalancerDriver(
        test_db_loadbalancerv2.LbaasPluginDbTestCase):

    def _create_fake_models(self):
        id = 'name-001'
        lb = data_models.LoadBalancer(id=id)
        listener = data_models.Listener(id=id, loadbalancer=lb)
        pool = data_models.Pool(id=id, loadbalancer=lb)
        member = data_models.Member(id=id, pool=pool)
        hm = data_models.HealthMonitor(id=id, pool=pool)
        lb.listeners = [listener]
        lb.pools = [pool]
        listener.default_pool = pool
        pool.members = [member]
        pool.healthmonitor = hm
        return lb

    def setUp(self):
        super(TestBrocadeLoadBalancerDriver, self).setUp()
        self.context = context.get_admin_context()
        self.plugin = mock.Mock()
        self.driver = driver.BrocadeLoadBalancerDriver(self.plugin)
        self.lb = self._create_fake_models()

    def test_load_balancer_ops(self):
        LoadBalancerManagerTest(self, self.driver.load_balancer,
                                self.lb)
        self.driver.device_driver.create_loadbalancer \
            .assert_called_once_with(self.lb)
        self.driver.device_driver.update_loadbalancer \
            .assert_called_once_with(self.lb, self.lb)
        self.driver.device_driver.delete_loadbalancer \
            .assert_called_once_with(self.lb)

    def test_listener_ops(self):
        ManagerTest(self, self.driver.listener, self.lb.listeners[0])
        self.driver.device_driver.create_listener \
            .assert_called_once_with(self.lb.listeners[0])
        self.driver.device_driver.update_listener \
            .assert_called_once_with(self.lb.listeners[0],
                                     self.lb.listeners[0])
        self.driver.device_driver.delete_listener \
            .assert_called_once_with(self.lb.listeners[0])

    def test_pool_ops(self):
        pool_fake_model = self.lb.listeners[0].default_pool
        ManagerTest(self, self.driver.pool,
                    pool_fake_model)
        self.driver.device_driver.update_pool \
            .assert_called_once_with(pool_fake_model, pool_fake_model)
        self.driver.device_driver.delete_pool \
            .assert_called_once_with(pool_fake_model)

    def test_member_ops(self):
        member_fake_model = self.lb.listeners[0].default_pool.members[0]
        ManagerTest(self, self.driver.member,
                    member_fake_model)
        self.driver.device_driver.create_member \
            .assert_called_once_with(member_fake_model)
        self.driver.device_driver.update_member \
            .assert_called_once_with(member_fake_model, member_fake_model)
        self.driver.device_driver.delete_member \
            .assert_called_once_with(member_fake_model)

    def test_health_monitor_ops(self):
        hm_fake_model = self.lb.listeners[0].default_pool.healthmonitor
        ManagerTest(self, self.driver.health_monitor, hm_fake_model)
        self.driver.device_driver.update_healthmonitor \
            .assert_called_once_with(hm_fake_model, hm_fake_model)
        self.driver.device_driver.delete_healthmonitor \
            .assert_called_once_with(hm_fake_model)



# Copyright 2015, Doug Wiegley (dougwig), A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import a10_neutron_lbaas
from oslo_log import log as logging

from neutron_lbaas.drivers import driver_base

VERSION = "2.0.0"
LOG = logging.getLogger(__name__)


class ThunderDriver(driver_base.LoadBalancerBaseDriver):

    def __init__(self, plugin):
        super(ThunderDriver, self).__init__(plugin)

        self.load_balancer = LoadBalancerManager(self)
        self.listener = ListenerManager(self)
        self.pool = PoolManager(self)
        self.member = MemberManager(self)
        self.health_monitor = HealthMonitorManager(self)

        LOG.debug("A10Driver: v2 initializing, version=%s, lbaas_manager=%s",
                  VERSION, a10_neutron_lbaas.VERSION)

        self.a10 = a10_neutron_lbaas.A10OpenstackLBV2(self)


class LoadBalancerManager(driver_base.BaseLoadBalancerManager):

    def create(self, context, lb):
        self.driver.a10.lb.create(context, lb)

    def update(self, context, old_lb, lb):
        self.driver.a10.lb.update(context, old_lb, lb)

    def delete(self, context, lb):
        self.driver.a10.lb.delete(context, lb)

    def refresh(self, context, lb):
        self.driver.a10.lb.refresh(context, lb)

    def stats(self, context, lb):
        return self.driver.a10.lb.stats(context, lb)


class ListenerManager(driver_base.BaseListenerManager):

    def create(self, context, listener):
        self.driver.a10.listener.create(context, listener)

    def update(self, context, old_listener, listener):
        self.driver.a10.listener.update(context, old_listener, listener)

    def delete(self, context, listener):
        self.driver.a10.listener.delete(context, listener)


class PoolManager(driver_base.BasePoolManager):

    def create(self, context, pool):
        self.driver.a10.pool.create(context, pool)

    def update(self, context, old_pool, pool):
        self.driver.a10.pool.update(context, old_pool, pool)

    def delete(self, context, pool):
        self.driver.a10.pool.delete(context, pool)


class MemberManager(driver_base.BaseMemberManager):

    def create(self, context, member):
        self.driver.a10.member.create(context, member)

    def update(self, context, old_member, member):
        self.driver.a10.member.update(context, old_member, member)

    def delete(self, context, member):
        self.driver.a10.member.delete(context, member)


class HealthMonitorManager(driver_base.BaseHealthMonitorManager):

    def create(self, context, hm):
        self.driver.a10.hm.create(context, hm)

    def update(self, context, old_hm, hm):
        self.driver.a10.hm.update(context, old_hm, hm)

    def delete(self, context, hm):
        self.driver.a10.hm.delete(context, hm)



# Copyright 2016 Rackspace
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lbaas._i18n import _
from oslo_config import cfg
from oslo_log import log as logging
import oslo_messaging as messaging
from oslo_messaging.rpc import dispatcher
from oslo_service import service


oslo_messaging_opts = [
    cfg.StrOpt('event_stream_topic',
               default='neutron_lbaas_event',
               deprecated_for_removal=True,
               deprecated_since='Queens',
               deprecated_reason='The neutron-lbaas project is now '
                                 'deprecated. See: https://wiki.openstack.org/'
                                 'wiki/Neutron/LBaaS/Deprecation',
               help=_('topic name for receiving events from a queue'))
]

cfg.CONF.register_opts(oslo_messaging_opts, group='oslo_messaging')


LOG = logging.getLogger(__name__)


class InfoContainer(object):
    @staticmethod
    def from_dict(dict_obj):
        return InfoContainer(dict_obj['info_type'],
                             dict_obj['info_id'],
                             dict_obj['info_payload'])

    def __init__(self, info_type, info_id, info_payload):
        self.info_type = info_type
        self.info_id = info_id
        self.info_payload = info_payload

    def to_dict(self):
        return {'info_type': self.info_type,
                'info_id': self.info_id,
                'info_payload': self.info_payload}

    def __eq__(self, other):
        if not isinstance(other, InfoContainer):
            return False
        if self.info_type != other.info_type:
            return False
        if self.info_id != other.info_id:
            return False
        if self.info_payload != other.info_payload:
            return False
        return True

    def __ne__(self, other):
        return not self == other


class ConsumerEndPoint(object):
    target = messaging.Target(namespace="control", version='1.0')

    def __init__(self, driver):
        self.driver = driver

    def update_info(self, ctx, container):
        LOG.debug("Received event from stream %s", container)
        container_inst = InfoContainer.from_dict(container)
        self.driver.handle_streamed_event(container_inst)


class OctaviaConsumer(service.Service):
    def __init__(self, driver, **kwargs):
        super(OctaviaConsumer, self).__init__(**kwargs)
        topic = cfg.CONF.oslo_messaging.event_stream_topic
        server = cfg.CONF.host
        self.driver = driver
        self.transport = messaging.get_rpc_transport(cfg.CONF)
        self.target = messaging.Target(topic=topic, server=server,
                                       exchange="common", fanout=False)
        self.endpoints = [ConsumerEndPoint(self.driver)]
        self.server = None

    def start(self):
        super(OctaviaConsumer, self).start()
        LOG.info("Starting octavia consumer...")
        access_policy = dispatcher.DefaultRPCAccessPolicy
        self.server = messaging.get_rpc_server(self.transport, self.target,
                                               self.endpoints,
                                               executor='eventlet',
                                               access_policy=access_policy)
        self.server.start()

    def stop(self, graceful=False):
        if self.server:
            LOG.info('Stopping consumer...')
            self.server.stop()
            if graceful:
                LOG.info(
                    ('Consumer successfully stopped.  Waiting for final '
                     'messages to be processed...'))
                self.server.wait()
        super(OctaviaConsumer, self).stop(graceful=graceful)

    def reset(self):
        if self.server:
            self.server.reset()
        super(OctaviaConsumer, self).reset()
